"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional

from encommon.types import BaseModel

from pydantic import Field

from .common import StatusPluginStates
from ...robie.params import RobiePluginParams
from ...robie.params.common import RobieParamsModel



class StatusPluginReportParams(BaseModel, extra='forbid'):
    """
    Contain information for constructing the chat messages.
    """

    client: Annotated[
        str,
        Field(...,
              description='Client where channel exists',
              min_length=1)]

    target: Annotated[
        str,
        Field(...,
              description='Where the message will be sent',
              min_length=1)]

    states: Annotated[
        Optional[list[StatusPluginStates]],
        Field(None,
              description='Which status value are related',
              min_length=1)]


    def __init__(
        # NOCVR
        self,
        /,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        states = data.get('states')

        if isinstance(states, str):
            data['states'] = [states]

        super().__init__(**data)



class StatusPluginCommandParams(RobieParamsModel, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    irc: Annotated[
        str,
        Field('!status',
              description='Command name for chat platform')]

    dsc: Annotated[
        str,
        Field('!status',
              description='Command name for chat platform')]

    mtm: Annotated[
        str,
        Field('!status',
              description='Command name for chat platform')]



class StatusPluginIconParams(BaseModel, extra='forbid'):
    """
    Contain information for constructing the chat messages.
    """

    irc: Annotated[
        Optional[str],
        Field(None,
              description='Icon used for the chat platform',
              min_length=1)]

    dsc: Annotated[
        Optional[str],
        Field(None,
              description='Icon used for the chat platform',
              min_length=1)]

    mtm: Annotated[
        Optional[str],
        Field(None,
              description='Icon used for the chat platform',
              min_length=1)]



class StatusPluginIconsParams(BaseModel, extra='forbid'):
    """
    Contain information for constructing the chat messages.
    """

    pending: Annotated[
        StatusPluginIconParams,
        Field(default_factory=StatusPluginIconParams,
              description='Icon used per the chat platform')]

    normal: Annotated[
        StatusPluginIconParams,
        Field(default_factory=StatusPluginIconParams,
              description='Icon used per the chat platform')]

    failure: Annotated[
        StatusPluginIconParams,
        Field(default_factory=StatusPluginIconParams,
              description='Icon used per the chat platform')]

    unknown: Annotated[
        StatusPluginIconParams,
        Field(default_factory=StatusPluginIconParams,
              description='Icon used per the chat platform')]



class StatusPluginParams(RobiePluginParams, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    reports: Annotated[
        Optional[list[StatusPluginReportParams]],
        Field(None,
              description='Where to send status updates',
              min_length=1)]

    command: Annotated[
        StatusPluginCommandParams,
        Field(default_factory=StatusPluginCommandParams,
              description='Command name per chat platform')]

    icons: Annotated[
        StatusPluginIconsParams,
        Field(default_factory=StatusPluginIconsParams,
              description='Icon used per the chat platform')]


    def __init__(
        # NOCVR
        self,
        /,
        _parse: Optional[Callable[..., Any]] = None,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        if _parse is not None:

            parsable = [
                'report',
                'command',
                'icons']

            for key in parsable:

                value = data.get(key)

                if value is None:
                    continue

                value = _parse(value)

                data[key] = value


        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Literal
from typing import TYPE_CHECKING

from encommon.times import Time

if TYPE_CHECKING:
    from .params import StatusPluginIconParams



StatusPluginStates = Literal[
    'pending',
    'normal',
    'failure',
    'unknown']



@dataclass
class StatusPluginItem:
    """
    Contain the relevant status information for the entry.
    """

    time: Time
    unique: str
    group: str
    title: str
    icon: 'StatusPluginIconParams'
    state: StatusPluginStates


    def __lt__(
        self,
        other: 'StatusPluginItem',
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        .. note::
           Useful with sorting to influence consistent output.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        unique = self.unique
        _unique = other.unique

        return unique < _unique



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .common import StatusPluginItem
from .common import StatusPluginStates
from .params import StatusPluginIconParams
from .params import StatusPluginParams
from .plugin import StatusPlugin



__all__ = [
    'StatusPlugin',
    'StatusPluginParams',
    'StatusPluginIconParams',
    'StatusPluginStates',
    'StatusPluginItem']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from .member import RobieMember
from ..threads import RobiePluginThread

if TYPE_CHECKING:
    from .member import RobieThreads



class RobiePlugins(RobieMember):
    """
    Common methods and routines for Chatting Robie members.
    """


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        vacate = self.vacate
        mqueue = self.mqueue

        threads = (
            self.threads
            .values())


        def _put_message() -> None:

            (thread
             .mqueue.put(mitem))


        while not mqueue.empty:

            mitem = mqueue.get()

            if vacate.is_set():
                continue

            for thread in threads:
                _put_message()


    def build_threads(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.

        .. note::
           Deviates from enhomie in build happens downstream,
           because all of the enhomie members are origin based
           where enrobie children do not have consistent base.
        """

        robie = self.robie
        childs = robie.childs
        plugins = childs.plugins

        model = RobiePluginThread


        threads: 'RobieThreads' = {}


        items = plugins.items()

        for name, plugin in items:

            object = model(
                self, plugin)

            threads[name] = object


        self.threads = threads



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from threading import enumerate as thread_enumerate
from time import sleep as block_sleep
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.types import NCNone

from enconnect.irc import Client
from enconnect.irc import ClientEvent

from .command import IRCCommand
from .message import IRCMessage
from .params import IRCClientParams
from ...plugins import StatusPlugin
from ...plugins import StatusPluginStates
from ...robie.addons import RobieQueue
from ...robie.childs import RobieClient
from ...utils import DupliThread

if TYPE_CHECKING:
    from ...robie.models import RobieCommand
    from ...robie.models import RobieMessage
    from ...robie.threads import RobieThread



class IRCClient(RobieClient):
    """
    Establish and maintain connection with the chat service.
    """

    __client: Client


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        params = self.params

        assert isinstance(
            params, IRCClientParams)

        client = Client(
            params.client,
            self.__debugger)

        self.__client = client

        self.__status('pending')


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Nothing to do for client


    @property
    def family(
        self,
    ) -> Literal['irc']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'irc'


    @property
    def client(
        self,
    ) -> Client:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    def operate(  # noqa: CFQ001
        self,
        thread: 'RobieThread',
    ) -> None:
        """
        Perform the operation related to Homie service threads.

        :param thread: Child class instance for Chatting Robie.
        """

        robie = thread.robie
        member = thread.member
        target = member.mqueue
        cqueue = thread.cqueue
        vacate = member.vacate
        params = self.params

        assert isinstance(
            params, IRCClientParams)

        delay = params.delay


        def _put_mqueue() -> None:

            event = source.get()

            self.put_message(
                target, event)


        def _get_cqueue() -> None:

            if not client.connected:
                return NCNone

            citem = cqueue.get()

            assert isinstance(
                citem, IRCCommand)

            try:
                client.socket_send(
                    citem.event)

            except Exception as reason:

                robie.logger.log_e(
                    base=self,
                    name=self,
                    status='exception',
                    exc_info=reason)


        def _continue() -> bool:

            return (
                not vacate.is_set()
                and daerht.is_alive())


        client = self.__client
        source = client.mqueue


        def _operate() -> None:

            try:
                client.operate()

            except ConnectionError:
                self.__status('pending')

            except Exception as reason:

                robie.logger.log_e(
                    base=self,
                    name=self,
                    status='exception',
                    exc_info=reason)

                self.__status('failure')


        def _routine() -> None:

            while _continue():

                robie.logger.log_i(
                    base=self,
                    name=self,
                    status='connect')

                self.__status('normal')

                _operate()

                robie.logger.log_i(
                    base=self,
                    name=self,
                    status='severed')

                block_sleep(delay)


        name = (
            f'{self.name}'
            '_thread_client')

        threads = (
            x.name for x in
            thread_enumerate())

        if name in threads:
            raise DupliThread(name)

        daerht = Thread(
            name=name,
            target=_routine)

        daerht.start()


        while _continue():

            if not cqueue.empty:
                _get_cqueue()

            block_sleep(0.025)

            while not source.empty():
                _put_mqueue()

            block_sleep(0.025)


        client.stop()

        while daerht.is_alive():
            daerht.join(1)

        daerht.join()

        while not source.empty():
            _put_mqueue()  # NOCVR


    def get_message(
        self,
        event: Optional[ClientEvent] = None,
    ) -> IRCMessage:
        """
        Return the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param event: Raw event received from the network peer.
        :returns: New item containing information for operation.
        """

        assert event is not None

        return IRCMessage(self, event)


    def put_message(
        self,
        mqueue: RobieQueue['RobieMessage'],
        event: Optional[ClientEvent] = None,
    ) -> None:
        """
        Insert the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param mqueue: Queue instance where the item is received.
        :param event: Raw event received from the network peer.
        """

        mitem = self.get_message(event)

        mqueue.put(mitem)


    def get_command(
        self,
        event: Optional[str] = None,
    ) -> IRCCommand:
        """
        Return the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param event: Raw event to be sent to the network peer.
        :returns: New item containing information for operation.
        """

        assert event is not None

        return IRCCommand(self, event)


    def put_command(
        self,
        cqueue: RobieQueue['RobieCommand'],
        event: Optional[str] = None,
    ) -> None:
        """
        Insert the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param cqueue: Queue instance where the item is received.
        :param event: Raw event to be sent to the network peer.
        """

        citem = self.get_command(event)

        cqueue.put(citem)


    def compose(
        self,
        target: str,
        content: str,
    ) -> 'RobieCommand':
        """
        Compose the message and transmit using the chat client.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param target: Where the composed message will be sent.
        :param content: Content that will be source of message.
        """

        assert target is not None

        return (
            self.get_command(
                f'PRIVMSG {target}'
                f' :{content}'))


    def __debugger(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        :param kwargs: Keyword arguments for populating message.
        """

        robie = self.robie

        assert not any([
            kwargs.get('level'),
            kwargs.get('base'),
            kwargs.get('name')])

        robie.logger.log_d(
            base=self,
            name=self,
            **kwargs)


    def __status(
        self,
        status: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param status: One of several possible value for status.
        """

        robie = self.robie
        childs = robie.childs
        plugins = childs.plugins
        params = self.params

        assert isinstance(
            params, IRCClientParams)

        if 'status' not in plugins:
            return None

        plugin = plugins['status']

        assert isinstance(
            plugin, StatusPlugin)

        (plugin.update(
            unique=self.name,
            group='Connections',
            title='IRC',
            icon=params.status,
            state=status))



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Literal
from typing import Optional
from typing import Type

from encommon.config import Config
from encommon.config import Params
from encommon.parse import Jinja2
from encommon.types import DictStrAny
from encommon.types import merge_dicts
from encommon.types import setate
from encommon.types import sort_dict
from encommon.utils.common import PATHABLE

from .params import RobieClientParams
from .params import RobieParams
from .params import RobiePluginParams



RobieConfigInsert = Literal[
    'clients', 'plugins']

RobieConfigInserts = dict[
    RobieConfigInsert, DictStrAny]



class RobieConfig(Config):
    """
    Contain the configurations from the arguments and files.

    :param sargs: Additional arguments on the command line.
    :param files: Complete or relative path to config files.
    :param cargs: Configuration arguments in dictionary form,
        which will override contents from the config files.
    """

    __clients: dict[str, Type[RobieClientParams]]
    __plugins: dict[str, Type[RobiePluginParams]]

    __inserts: RobieConfigInserts


    def __init__(
        self,
        sargs: Optional[DictStrAny] = None,
        files: Optional[PATHABLE] = None,
        cargs: Optional[DictStrAny] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        sargs = dict(sargs or {})
        cargs = dict(cargs or {})

        self.__clients = {}
        self.__plugins = {}

        self.__inserts = {}


        _console = (
            sargs.get('console'))

        _debug = (
            sargs.get('debug'))

        key = 'enlogger/stdo_level'

        if _console is True:
            cargs[key] = 'info'

        if _debug is True:
            cargs[key] = 'debug'


        if 'config' in sargs:
            files = sargs['config']


        _pmessage = (
            sargs.get('pmessage'))

        _pcommand = (
            sargs.get('pcommand'))

        if _pmessage is not None:
            key = 'printer/message'
            cargs[key] = _pmessage

        if _pcommand is not None:
            key = 'printer/command'
            cargs[key] = _pcommand


        super().__init__(
            files=files,
            cargs=cargs,
            sargs=sargs,
            model=RobieParams)

        self.merge_params()


    @property
    def inserts(
        self,
    ) -> RobieConfigInserts:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return deepcopy(self.__inserts)


    @property
    def params(
        self,
    ) -> RobieParams:
        """
        Return the Pydantic model containing the configuration.

        .. warning::
           This method completely overrides the parent but is
           based on that code, would be unfortunate if upstream
           changes meant this breaks or breaks something else.

        :returns: Pydantic model containing the configuration.
        """

        params = self.__params

        if params is not None:

            assert isinstance(
                params, RobieParams)

            return params


        basic = self.basic

        enconfig = (
            basic.get('enconfig'))

        enlogger = (
            basic.get('enlogger'))

        encrypts = (
            basic.get('encrypts'))

        basic = {
            'enconfig': enconfig,
            'enlogger': enlogger,
            'encrypts': encrypts}

        params = (
            self.model(**basic))

        assert isinstance(
            params, RobieParams)


        self.__params = params

        return self.__params


    def merge_params(
        self,
    ) -> None:
        """
        Update the Pydantic model containing the configuration.
        """

        fclients: DictStrAny = {}
        fplugins: DictStrAny = {}


        merge = self.merge

        merge_dicts(
            dict1=merge,
            dict2=self.inserts,
            force=True)


        _clients = merge.get('clients')
        _plugins = merge.get('plugins')


        jinja2 = Jinja2({
            'source': merge,
            'config': self})

        parse = jinja2.parse


        def _put_clients() -> None:

            models = self.__clients

            items = (
                merge['clients']
                .items())

            for name, params in items:

                if name not in models:
                    continue  # NOCVR

                model = models[name]

                object = model(
                    parse, **params)

                fclients[name] = object


        if _clients is not None:

            _put_clients()

            del merge['clients']


        def _put_plugins() -> None:

            models = self.__plugins

            items = (
                merge['plugins']
                .items())

            for name, params in items:

                if name not in models:
                    continue  # NOCVR

                model = models[name]

                object = model(
                    parse, **params)

                fplugins[name] = object


        if _plugins is not None:

            _put_plugins()

            del merge['plugins']


        if len(fclients) >= 1:
            merge['clients'] = fclients

        if len(fplugins) >= 1:
            merge['plugins'] = fplugins


        params = self.model(
            parse, **merge)

        assert isinstance(
            params, RobieParams)


        self.__params = params


    @property
    def __params(
        self,
    ) -> Optional[Params]:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self._Config__params


    @__params.setter
    def __params(
        self,
        value: Params,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        self._Config__params = value


    @property
    def config(
        self,
    ) -> DictStrAny:
        """
        Return the configuration dumped from the Pydantic model.

        .. warning::
           This method completely overrides the parent but is
           based on that code, would be unfortunate if upstream
           changes meant this breaks or breaks something else.

        :returns: Configuration dumped from the Pydantic model.
        """

        params = self.params

        _clients = params.clients
        _plugins = params.plugins

        config = params.endumped


        def _get_clients() -> None:

            target = config['clients']

            assert _clients is not None

            items = _clients.items()

            for name, client in items:

                target[name] = (
                    client.endumped)


        if _clients is not None:
            _get_clients()


        def _get_plugins() -> None:

            target = config['plugins']

            assert _plugins is not None

            items = _plugins.items()

            for name, plugin in items:

                target[name] = (
                    plugin.endumped)


        if _plugins is not None:
            _get_plugins()


        return sort_dict(config)


    def register(
        self,
        name: str,
        *,
        client: Optional[Type[RobieClientParams]] = None,
        plugin: Optional[Type[RobiePluginParams]] = None,
        source: Optional[DictStrAny] = None,
        merge: bool = True,
    ) -> None:
        """
        Register the plugin parameters for parameter processing.

        :param name: Name of the object within the Robie config.
        :param client: Class definition for the instantiation.
        :param plugin: Class definition for the instantiation.
        :param source: Source for the parameters instantiation.
        :param merge: Reprocess all parameters including added.
        """

        clients = self.__clients
        plugins = self.__plugins
        inserts = self.__inserts


        def _insert(
            target: RobieConfigInsert,
        ) -> None:

            if source is None:
                return None

            setate(  # NOCVR
                inserts,  # type: ignore
                f'{target}/{name}',
                source)


        if client is not None:

            clients[name] = client

            _insert('clients')


        if plugin is not None:

            plugins[name] = plugin

            _insert('plugins')


        if merge is True:
            self.merge_params()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING
from typing import Type

if TYPE_CHECKING:
    from ..addons import RobieQueueItem
    from ..models import RobieCommand
    from ..models import RobieMessage
    from ..params import RobieParams
    from ..params import RobiePrinterParams
    from ..params import RobieServiceParams
    from ..params import RobieChildParams
    from ..params import RobieClientParams
    from ..params import RobiePluginParams



class RobieModels:
    """
    Return the class object that was imported within method.
    """


    @classmethod
    def robie(
        cls,
    ) -> Type['RobieParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..params import (
            RobieParams)

        return RobieParams


    @classmethod
    def printer(
        cls,
    ) -> Type['RobiePrinterParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..params import (
            RobiePrinterParams)

        return RobiePrinterParams


    @classmethod
    def service(
        cls,
    ) -> Type['RobieServiceParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..params import (
            RobieServiceParams)

        return RobieServiceParams


    @classmethod
    def child(
        cls,
    ) -> Type['RobieChildParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..params import (
            RobieChildParams)

        return RobieChildParams


    @classmethod
    def client(
        cls,
    ) -> Type['RobieClientParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..params import (
            RobieClientParams)

        return RobieClientParams


    @classmethod
    def plugin(
        cls,
    ) -> Type['RobiePluginParams']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..params import (
            RobiePluginParams)

        return RobiePluginParams


    @classmethod
    def queue(
        cls,
    ) -> Type['RobieQueueItem']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..addons import (
            RobieQueueItem)

        return RobieQueueItem


    @classmethod
    def message(
        cls,
    ) -> Type['RobieMessage']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..models import (
            RobieMessage)

        return RobieMessage


    @classmethod
    def command(
        cls,
    ) -> Type['RobieCommand']:
        """
        Return the class object that was imported within method.

        :returns: Class object that was imported within method.
        """

        from ..models import (
            RobieCommand)

        return RobieCommand



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal

from .child import RobieChild



class RobiePlugin(RobieChild):
    """
    Integrate with the Robie routine and perform operations.
    """


    @property
    def kind(
        self,
    ) -> Literal['plugin']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'plugin'



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...robie import Robie



def test_RobieChilds(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs


    attrs = lattrs(childs)

    assert attrs == [
        '_RobieChilds__robie',
        '_RobieChilds__clients',
        '_RobieChilds__plugins']


    assert inrepr(
        'robie.RobieChilds',
        childs)

    assert isinstance(
        hash(childs), int)

    assert instr(
        'robie.RobieChilds',
        childs)


    childs.validate()

    assert childs.clients

    assert childs.plugins



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from argparse import ArgumentParser
from signal import SIGHUP
from signal import SIGINT
from signal import SIGTERM
from signal import signal
from sys import argv as sys_argv
from typing import Optional

from encommon.types import DictStrAny

from ..clients import DSCClient
from ..clients import DSCClientParams
from ..clients import IRCClient
from ..clients import IRCClientParams
from ..clients import MTMClient
from ..clients import MTMClientParams
from ..plugins import StatusPlugin
from ..plugins import StatusPluginParams
from ..robie import Robie
from ..robie import RobieConfig
from ..robie import RobieService



def arguments(
    args: Optional[list[str]] = None,
) -> DictStrAny:
    """
    Construct arguments which are associated with the file.

    :param args: Override the source for the main arguments.
    :returns: Construct arguments from command line options.
    """

    parser = ArgumentParser()

    args = args or sys_argv[1:]


    parser.add_argument(
        '--config',
        help=(
            'complete or relative '
            'path to config file'))


    parser.add_argument(
        '--console',
        action='store_true',
        default=False,
        help=(
            'write log messages '
            'to standard output'))


    parser.add_argument(
        '--debug',
        action='store_true',
        default=False,
        help=(
            'increase logging level '
            'for standard output'))


    parser.add_argument(
        '--print_message',
        action='store_true',
        default=False,
        dest='pmessage',
        help=(
            'print the received '
            'messages to console'))


    parser.add_argument(
        '--print_command',
        action='store_true',
        default=False,
        dest='pcommand',
        help=(
            'print the submited '
            'commands to console'))


    return vars(
        parser
        .parse_args(args))



def operation(
    # NOCVR
    robie: Robie,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param robie: Primary class instance for Chatting Robie.
    """

    service = RobieService(robie)

    service.start()

    signal(SIGINT, service.soft)
    signal(SIGTERM, service.soft)
    signal(SIGHUP, service.soft)

    service.operate()



def register_params(
    # NOCVR
    config: RobieConfig,
) -> None:
    """
    Register the plugin parameters for parameter processing.

    :param config: Primary class instance for configuration.
    """

    config.register(
        name='ircbot',
        client=IRCClientParams)

    config.register(
        name='dscbot',
        client=DSCClientParams)

    config.register(
        name='mtmbot',
        client=MTMClientParams)

    config.register(
        name='status',
        plugin=StatusPluginParams)



def register_plugins(
    # NOCVR
    robie: Robie,
) -> None:
    """
    Register the plugin parameters for parameter processing.

    :param robie: Primary class instance for Chatting Robie.
    """

    robie.register(
        name='ircbot',
        client=IRCClient)

    robie.register(
        name='dscbot',
        client=DSCClient)

    robie.register(
        name='mtmbot',
        client=MTMClient)

    robie.register(
        name='status',
        plugin=StatusPlugin)



def execution(
    # NOCVR
    args: Optional[list[str]] = None,
) -> None:
    """
    Perform whatever operation is associated with the file.

    :param args: Override the source for the main arguments.
    """

    config = RobieConfig(
        arguments(args))

    register_params(config)

    config.logger.start()

    config.logger.log_i(
        base='execution/service',
        status='started')

    robie = Robie(config)

    register_plugins(robie)

    operation(robie)

    config.logger.log_i(
        base='execution/service',
        status='stopped')

    config.logger.stop()



if __name__ == '__main__':
    execution()  # NOCVR



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from pytest import raises

from ..command import MTMCommand

if TYPE_CHECKING:
    from ....robie import Robie



def test_MTMCommand(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    model = MTMCommand

    client = clients['mtmbot']


    item = model(
        client, 'delete',
        'posts/mocked')


    attrs = lattrs(item)

    assert attrs == [
        'method',
        'path',
        'params',
        'json',
        'client',
        'time']


    assert inrepr(
        'MTMCommand',
        item)

    with raises(TypeError):
        hash(item)

    assert instr(
        'MTMCommand',
        item)


    assert item.time.since > 0

    assert item.client == client.name

    assert item.family == 'mattermost'

    assert item.method == 'delete'

    assert item.path == 'posts/mocked'

    assert not item.params

    assert not item.json



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .status import StatusPlugin
from .status import StatusPluginIconParams
from .status import StatusPluginItem
from .status import StatusPluginParams
from .status import StatusPluginStates



__all__ = [
    'StatusPlugin',
    'StatusPluginParams',
    'StatusPluginIconParams',
    'StatusPluginItem',
    'StatusPluginStates']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .thread import RobieThread



class RobieClientThread(RobieThread):
    """
    Common methods and routines for Chatting Robie threads.
    """



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.times import Duration
from encommon.types import DictStrAny
from encommon.types import sort_dict
from encommon.types.strings import NEWLINE
from encommon.types.strings import SEMPTY

from .params import StatusPluginParams

if TYPE_CHECKING:
    from ...robie.addons import RobieQueue
    from ...robie.childs import RobieClient
    from ...robie.childs import RobiePlugin
    from ...robie.models import RobieCommand
    from ...robie.models import RobieMessage
    from .common import StatusPluginItem
    from .params import StatusPluginReportParams
    from .plugin import StatusPlugin



def grouped(
    status: dict[str, 'StatusPluginItem'],
) -> dict[str, list['StatusPluginItem']]:
    """
    Return the dictionary with status value stored by group.

    :param status: Object containing the status information.
    :returns: Dictionary with status value stored by group.
    """

    groups: DictStrAny = {
        x.group: [] for x
        in status.values()}

    items = status.items()

    for name, value in items:

        group = value.group

        (groups[group]
         .append(value))

    return sort_dict(groups)



def composedsc(
    plugin: 'StatusPlugin',
    cqueue: 'RobieQueue[RobieCommand]',
    mitem: 'RobieMessage',
    status: dict[str, 'StatusPluginItem'],
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param cqueue: Queue instance where the item is received.
    :param mitem: Item containing information for operation.
    :param status: Object containing the status information.
    """

    robie = plugin.robie
    params = plugin.params

    compose: list[str] = []


    def _compose(
        message: str,
    ) -> None:

        citem = mitem.reply(
            robie, message)

        cqueue.put(citem)


    def _format() -> None:

        assert isinstance(
            params,
            StatusPluginParams)

        epoch = int(value.time)

        sicon = (
            (getattr(
                params.icons,
                value.state)
             .dsc)
            or SEMPTY)

        sicon = (
            f'{sicon} '
            if len(sicon) >= 1
            else sicon)

        vicon = (
            value.icon.dsc
            or SEMPTY)

        compose.append(
            (f'{vicon or ""} '
             f'{value.title}/'
             f'`{value.unique}`:'
             f' {sicon or ""}'
             f'`{value.state}` '
             f'<t:{epoch}:R>')
            .strip())


    items = sorted(
        grouped(status)
        .items())

    for name, values in items:

        compose.append(
            f'**{name}**')

        _values = sorted(values)

        for value in _values:
            _format()

    _compose(
        NEWLINE
        .join(compose))



def reportdsc(
    plugin: 'RobiePlugin',
    client: 'RobieClient',
    cqueue: 'RobieQueue[RobieCommand]',
    status: 'StatusPluginItem',
    report: 'StatusPluginReportParams',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param client: Client class instance for Chatting Robie.
    :param cqueue: Queue instance where the item is received.
    :param status: Object containing the status information.
    :param report: Object containing the report information.
    """

    params = plugin.params

    assert isinstance(
        params,
        StatusPluginParams)


    def _compose(
        message: str,
    ) -> None:

        citem = (
            client.compose(
                report.target,
                message))

        cqueue.put(citem)


    epoch = int(status.time)

    sicon = (
        (getattr(
            params.icons,
            status.state)
         .dsc)
        or SEMPTY)

    sicon = (
        f'{sicon} '
        if len(sicon) >= 1
        else sicon)

    vicon = (
        status.icon.dsc
        or SEMPTY)

    _compose(
        (f'{vicon or ""} '
         f'{status.title}/'
         f'`{status.unique}`:'
         f' {sicon or ""}'
         f'`{status.state}` '
         f'<t:{epoch}:R>')
        .strip())



def composeirc(
    plugin: 'StatusPlugin',
    cqueue: 'RobieQueue[RobieCommand]',
    mitem: 'RobieMessage',
    status: dict[str, 'StatusPluginItem'],
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param cqueue: Queue instance where the item is received.
    :param mitem: Item containing information for operation.
    :param status: Object containing the status information.
    """

    robie = plugin.robie
    params = plugin.params


    def _compose(
        message: str,
    ) -> None:

        citem = mitem.reply(
            robie, message)

        cqueue.put(citem)


    def _format() -> None:

        assert isinstance(
            params,
            StatusPluginParams)

        durate = Duration(
            value.time.since)

        sicon = (
            (getattr(
                params.icons,
                value.state)
             .irc)
            or SEMPTY)

        sicon = (
            f'{sicon} '
            if len(sicon) >= 1
            else sicon)

        vicon = (
            value.icon.irc
            or SEMPTY)

        _compose(
            (f'{vicon or ""} '
             f'{value.title}/'
             f'{value.unique}:'
             f' {sicon or ""}'
             f'{value.state} '
             f'for {durate}')
            .strip())


    items = sorted(
        grouped(status)
        .items())

    for name, values in items:

        _compose(name)

        _values = sorted(values)

        for value in _values:
            _format()



def reportirc(
    plugin: 'RobiePlugin',
    client: 'RobieClient',
    cqueue: 'RobieQueue[RobieCommand]',
    status: 'StatusPluginItem',
    report: 'StatusPluginReportParams',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param client: Client class instance for Chatting Robie.
    :param cqueue: Queue instance where the item is received.
    :param status: Object containing the status information.
    :param report: Object containing the report information.
    """

    params = plugin.params

    assert isinstance(
        params,
        StatusPluginParams)


    def _compose(
        message: str,
    ) -> None:

        citem = (
            client.compose(
                report.target,
                message))

        cqueue.put(citem)


    durate = Duration(
        status.time.since)

    sicon = (
        (getattr(
            params.icons,
            status.state)
         .irc)
        or SEMPTY)

    sicon = (
        f'{sicon} '
        if len(sicon) >= 1
        else sicon)

    vicon = (
        status.icon.irc
        or SEMPTY)

    _compose(
        (f'{vicon or ""} '
         f'{status.title}/'
         f'{status.unique}:'
         f' {sicon or ""}'
         f'{status.state} '
         f'for {durate}')
        .strip())



def composemtm(
    plugin: 'StatusPlugin',
    cqueue: 'RobieQueue[RobieCommand]',
    mitem: 'RobieMessage',
    status: dict[str, 'StatusPluginItem'],
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param cqueue: Queue instance where the item is received.
    :param mitem: Item containing information for operation.
    :param status: Object containing the status information.
    """

    robie = plugin.robie
    params = plugin.params

    compose: list[str] = []


    def _compose(
        message: str,
    ) -> None:

        citem = mitem.reply(
            robie, message)

        cqueue.put(citem)


    def _format() -> None:

        assert isinstance(
            params,
            StatusPluginParams)

        durate = Duration(
            value.time.since)

        sicon = (
            (getattr(
                params.icons,
                value.state)
             .mtm)
            or SEMPTY)

        sicon = (
            f'{sicon} '
            if len(sicon) >= 1
            else sicon)

        vicon = (
            value.icon.mtm
            or SEMPTY)

        compose.append(
            (f'{vicon or ""} '
             f'{value.title}/'
             f'`{value.unique}`:'
             f' {sicon or ""}'
             f'`{value.state}` '
             f'for `{durate}`')
            .strip())


    items = sorted(
        grouped(status)
        .items())

    for name, values in items:

        compose.append(
            f'**{name}**')

        _values = sorted(values)

        for value in _values:
            _format()

    _compose(
        NEWLINE
        .join(compose))



def reportmtm(
    plugin: 'RobiePlugin',
    client: 'RobieClient',
    cqueue: 'RobieQueue[RobieCommand]',
    status: 'StatusPluginItem',
    report: 'StatusPluginReportParams',
) -> None:
    """
    Construct and format message for related chat platform.

    :param plugin: Plugin class instance for Chatting Robie.
    :param client: Client class instance for Chatting Robie.
    :param cqueue: Queue instance where the item is received.
    :param status: Object containing the status information.
    :param report: Object containing the report information.
    """

    params = plugin.params

    assert isinstance(
        params,
        StatusPluginParams)


    def _compose(
        message: str,
    ) -> None:

        citem = (
            client.compose(
                report.target,
                message))

        cqueue.put(citem)


    durate = Duration(
        status.time.since)

    sicon = (
        (getattr(
            params.icons,
            status.state)
         .mtm)
        or SEMPTY)

    sicon = (
        f'{sicon} '
        if len(sicon) >= 1
        else sicon)

    vicon = (
        status.icon.mtm
        or SEMPTY)

    _compose(
        (f'{vicon or ""} '
         f'{status.title}/'
         f'`{status.unique}`:'
         f' {sicon or ""}'
         f'`{status.state}` '
         f'for `{durate}`')
        .strip())



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Time
from encommon.types import NCNone

from .common import StatusPluginItem
from .common import StatusPluginStates
from .helpers import composedsc
from .helpers import composeirc
from .helpers import composemtm
from .helpers import reportdsc
from .helpers import reportirc
from .helpers import reportmtm
from .params import StatusPluginIconParams
from .params import StatusPluginParams
from ...robie.childs import RobiePlugin

if TYPE_CHECKING:
    from ...robie.threads import RobieThread



class StatusPlugin(RobiePlugin):
    """
    Integrate with the Robie routine and perform operations.

    .. note::
       This plugin responds to inquiries about Robie status.
    """

    __status: dict[str, StatusPluginItem]
    __report: dict[str, StatusPluginItem]


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__status = {}
        self.__report = {}


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        # Nothing to do for plugin


    def operate(
        self,
        thread: 'RobieThread',
    ) -> None:
        """
        Perform the operation related to Homie service threads.

        :param thread: Child class instance for Chatting Robie.
        """

        mqueue = thread.mqueue
        member = thread.member
        cqueue = member.cqueue
        params = self.params
        status = self.__status

        assert isinstance(
            params,
            StatusPluginParams)

        command = params.command
        match: Optional[str]

        kinds = ['privmsg', 'chanmsg']


        while not mqueue.empty:

            self.report(thread)

            mitem = mqueue.get()

            time = mitem.time
            kind = mitem.kind
            isme = mitem.isme
            family = mitem.family

            if kind not in kinds:
                continue

            if isme is True:
                continue

            if time.since > 15:
                continue  # NOCVR

            event = getattr(
                mitem, 'event')


            match = None

            if family == 'discord':
                match = command.dsc

            if family == 'irc':
                match = command.irc

            if family == 'mattermost':
                match = command.mtm

            if match is None:
                continue  # NOCVR


            message = event.message

            if message != match:
                continue  # NOCVR


            if family == 'discord':
                composedsc(
                    self, cqueue,
                    mitem, status)

            if family == 'irc':
                composeirc(
                    self, cqueue,
                    mitem, status)

            if family == 'mattermost':
                composemtm(
                    self, cqueue,
                    mitem, status)


    def report(
        self,
        thread: 'RobieThread',
    ) -> None:
        """
        Perform the operation related to Homie service threads.

        :param thread: Child class instance for Chatting Robie.
        """

        robie = self.robie
        childs = robie.childs
        clients = childs.clients
        member = thread.member
        cqueue = member.cqueue
        params = self.params
        status = self.__status
        report = self.__report

        assert isinstance(
            params,
            StatusPluginParams)

        reports = params.reports

        if reports is None:
            return NCNone


        items = status.items()

        for unique, _status in items:

            _report = (
                report.get(unique))

            if _report == _status:
                continue

            report[unique] = _status

            state = _status.state

            for item in reports:

                name = item.client
                states = item.states

                if (states is not None
                        and state not in states):
                    continue  # NOCVR

                if name not in clients:
                    continue  # NOCVR

                client = clients[name]

                family = client.family

                if family == 'discord':
                    reportdsc(
                        self, client,
                        cqueue,
                        _status, item)

                if family == 'irc':
                    reportirc(
                        self, client,
                        cqueue,
                        _status, item)

                if family == 'mattermost':
                    reportmtm(
                        self, client,
                        cqueue,
                        _status, item)


    def update(
        self,
        unique: str,
        group: str,
        title: str,
        icon: StatusPluginIconParams,
        state: StatusPluginStates,
    ) -> None:
        """
        Update or insert the status of the Robie child instance.

        :param unique: Unique identifier to use for the status.
        :param group: Name for the group the status is membered.
        :param title: Friendly name of the related unique entry.
        :param icon: Optional icon object if supported platform.
        :param state: One of several possible value for status.
        """

        robie = self.robie
        status = self.__status

        robie.logger.log_i(
            base=self,
            name=self,
            item='status',
            unique=unique,
            state=state)

        object = StatusPluginItem(
            unique=unique,
            time=Time(),
            group=group,
            title=title,
            icon=icon,
            state=state)

        status[unique] = object
